#===============================================================================
# Copyright 2007 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================


class LockError(Exception):
    pass


class Storage(object):
    """Abstract base class for storage objects.
    """
    
    def create_index(self, schema, indexname=None):
        raise NotImplementedError
    
    def open_index(self, indexname=None, schema=None):
        raise NotImplementedError
    
    def close(self):
        pass
    
    def optimize(self):
        pass










#===============================================================================
# Copyright 2009 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

import cPickle, re
from bisect import bisect_right
from time import time
from threading import Lock

from whoosh import __version__
from whoosh.fields import Schema
from whoosh.index import Index
from whoosh.index import EmptyIndexError, OutOfDateError, IndexVersionError
from whoosh.index import _DEF_INDEX_NAME
from whoosh.store import LockError
from whoosh.support.bitvector import BitVector
from whoosh.system import _INT_SIZE, _FLOAT_SIZE


_INDEX_VERSION = -105
_EXTENSIONS = "dci|dcz|tiz|fvz|pst|vps"


# A mix-in that adds methods for deleting
# documents from self.segments. These methods are on IndexWriter as
# well as Index for convenience, so they're broken out here.

class SegmentDeletionMixin(object):
    """Mix-in for classes that support deleting documents from self.segments."""
    
    def delete_document(self, docnum, delete = True):
        """Deletes a document by number."""
        self.segments.delete_document(docnum, delete = delete)
    
    def deleted_count(self):
        """Returns the total number of deleted documents in this index.
        """
        return self.segments.deleted_count()
    
    def is_deleted(self, docnum):
        """Returns True if a given document number is deleted but
        not yet optimized out of the index.
        """
        return self.segments.is_deleted(docnum)
    
    def has_deletions(self):
        """Returns True if this index has documents that are marked
        deleted but haven't been optimized out of the index yet.
        """
        return self.segments.has_deletions()
    

class FileIndex(SegmentDeletionMixin, Index):
    def __init__(self, storage, schema, create=False, indexname=_DEF_INDEX_NAME):
        self.storage = storage
        self.indexname = indexname
        
        if schema is not None and not isinstance(schema, Schema):
            raise ValueError("%r is not a Schema object" % schema)
        
        self.generation = self.latest_generation()
        
        if create:
            if schema is None:
                raise IndexError("To create an index you must specify a schema")
            
            self.schema = schema
            self.generation = 0
            self.segment_counter = 0
            self.segments = SegmentSet()
            
            # Clear existing files
            self.unlock()
            prefix = "_%s_" % self.indexname
            for filename in self.storage:
                if filename.startswith(prefix):
                    storage.delete_file(filename)
            
            self._write()
        elif self.generation >= 0:
            self._read(schema)
        else:
            raise EmptyIndexError
        
        # Open a reader for this index. This is used by the
        # deletion methods, but mostly it's to keep the underlying
        # files open so they don't get deleted from underneath us.
        self._searcher = self.searcher()
        
        self.segment_num_lock = Lock()
    
    def __repr__(self):
        return "%s(%r, %r)" % (self.__class__.__name__, self.storage, self.indexname)
    
    def __del__(self):
        if hasattr(self, "_searcher") and self._searcher and not self._searcher.is_closed:
            self._searcher.close()
            
    def close(self):
        self._searcher.close()
        
    def latest_generation(self):
        pattern = _toc_pattern(self.indexname)
        
        max = -1
        for filename in self.storage:
            m = pattern.match(filename)
            if m:
                num = int(m.group(1))
                if num > max: max = num
        return max
    
    def refresh(self):
        if not self.up_to_date():
            return self.__class__(self.storage, self.schema, indexname = self.indexname)
        else:
            return self
        
    def up_to_date(self):
        return self.generation == self.latest_generation()
    
    def _write(self):
        # Writes the content of this index to the .toc file.
        for field in self.schema:
            field.clean()
        #stream = self.storage.create_file(self._toc_filename())
        
        # Use a temporary file for atomic write.
        tocfilename = self._toc_filename()
        tempfilename = '%s.%s' % (tocfilename, time())
        stream = self.storage.create_file(tempfilename)
        
        stream.write_varint(_INT_SIZE)
        stream.write_varint(_FLOAT_SIZE)
        stream.write_int(-12345)
        
        stream.write_int(_INDEX_VERSION)
        for num in __version__[:3]:
            stream.write_varint(num)
        
        stream.write_string(cPickle.dumps(self.schema, -1))
        stream.write_int(self.generation)
        stream.write_int(self.segment_counter)
        stream.write_pickle(self.segments)
        stream.close()
        
        # Rename temporary file to the proper filename
        self.storage.rename_file(tempfilename, self._toc_filename(), safe=True)
    
    def _read(self, schema):
        # Reads the content of this index from the .toc file.
        stream = self.storage.open_file(self._toc_filename())
        
        if stream.read_varint() != _INT_SIZE or \
           stream.read_varint() != _FLOAT_SIZE:
            raise IndexError("Index was created on an architecture with different data sizes")
        
        if not stream.read_int() == -12345:
            raise IndexError("Number misread: byte order problem")
        
        version = stream.read_int()
        if version != _INDEX_VERSION:
            raise IndexVersionError("Can't read format %s" % version, version)
        self.version = version
        self.release = (stream.read_varint(),
                        stream.read_varint(),
                        stream.read_varint())
        
        # If the user supplied a schema object with the constructor,
        # don't load the pickled schema from the saved index.
        if schema:
            self.schema = schema
            stream.skip_string()
        else:
            self.schema = cPickle.loads(stream.read_string())
        
        generation = stream.read_int()
        assert generation == self.generation
        self.segment_counter = stream.read_int()
        self.segments = stream.read_pickle()
        stream.close()
    
    def _next_segment_name(self):
        #Returns the name of the next segment in sequence.
        if self.segment_num_lock.acquire():
            try:
                self.segment_counter += 1
                return "_%s_%s" % (self.indexname, self.segment_counter)
            finally:
                self.segment_num_lock.release()
        else:
            raise LockError
    
    def _toc_filename(self):
        # Returns the computed filename of the TOC for this
        # index name and generation.
        return "_%s_%s.toc" % (self.indexname, self.generation)
    
    def last_modified(self):
        return self.storage.file_modified(self._toc_filename())
    
    def lock(self):
        return self.storage.lock("_%s_LOCK" % self.indexname)
    
    def unlock(self):
        self.storage.unlock("_%s_LOCK" % self.indexname)
        
    def is_empty(self):
        return len(self.segments) == 0
    
    def optimize(self):
        if len(self.segments) < 2 and not self.segments.has_deletions():
            return
        
        from whoosh.filedb.filewriting import OPTIMIZE
        w = self.writer()
        w.commit(OPTIMIZE)
        
    def commit(self, new_segments = None):
        self._searcher.close()
        
        if not self.up_to_date():
            raise OutOfDateError
        
        if new_segments:
            self.segments = new_segments
        
        self.generation += 1
        self._write()
        self._clean_files()
        
        self._searcher = self.searcher()
        
    def _clean_files(self):
        # Attempts to remove unused index files (called when a new generation
        # is created). If existing Index and/or reader objects have the files
        # open, they may not get deleted immediately (i.e. on Windows)
        # but will probably be deleted eventually by a later call to clean_files.
        
        storage = self.storage
        current_segment_names = set([s.name for s in self.segments])
        
        tocpattern = _toc_pattern(self.indexname)
        segpattern = _segment_pattern(self.indexname)
        
        for filename in storage:
            m = tocpattern.match(filename)
            if m:
                num = int(m.group(1))
                if num != self.generation:
                    try:
                        storage.delete_file(filename)
                    except OSError:
                        # Another process still has this file open
                        pass
            else:
                m = segpattern.match(filename)
                if m:
                    name = m.group(1)
                    if name not in current_segment_names:
                        try:
                            storage.delete_file(filename)
                        except OSError:
                            # Another process still has this file open
                            pass
                        
    def doc_count_all(self):
        return self.segments.doc_count_all()
    
    def doc_count(self):
        return self.segments.doc_count()
    
    def field_length(self, fieldid):
        fieldnum = self.schema.to_number(fieldid)
        return sum(s.field_length(fieldnum) for s in self.segments)
    
    def reader(self):
        return self.segments.reader(self.storage, self.schema)
    
    def writer(self, **kwargs):
        from whoosh.filedb.filewriting import FileIndexWriter
        return FileIndexWriter(self, **kwargs)
        

# SegmentSet object

class SegmentSet(object):
    """This class is never instantiated by the user. It is used by the Index
    object to keep track of the segments in the index.
    """

    def __init__(self, segments = None):
        if segments is None:
            self.segments = []
        else:
            self.segments = segments
        
        self._doc_offsets = self.doc_offsets()
    
    def __repr__(self):
        return repr(self.segments)
    
    def __len__(self):
        """:returns: the number of segments in this set."""
        return len(self.segments)
    
    def __iter__(self):
        return iter(self.segments)
    
    def __getitem__(self, n):
        return self.segments.__getitem__(n)
    
    def append(self, segment):
        """Adds a segment to this set."""
        
        self.segments.append(segment)
        self._doc_offsets = self.doc_offsets()
    
    def _document_segment(self, docnum):
        """Returns the index.Segment object containing the given document
        number.
        """
        
        offsets = self._doc_offsets
        if len(offsets) == 1: return 0
        return bisect_right(offsets, docnum) - 1
    
    def _segment_and_docnum(self, docnum):
        """Returns an (index.Segment, segment_docnum) pair for the
        segment containing the given document number.
        """
        
        segmentnum = self._document_segment(docnum)
        offset = self._doc_offsets[segmentnum]
        segment = self.segments[segmentnum]
        return segment, docnum - offset
    
    def copy(self):
        """:returns: a deep copy of this set."""
        return self.__class__([s.copy() for s in self.segments])
    
    def doc_offsets(self):
        # Recomputes the document offset list. This must be called if you
        # change self.segments.
        offsets = []
        base = 0
        for s in self.segments:
            offsets.append(base)
            base += s.doc_count_all()
        return offsets
    
    def doc_count_all(self):
        """
        :returns: the total number of documents, DELETED or
            UNDELETED, in this set.
        """
        return sum(s.doc_count_all() for s in self.segments)
    
    def doc_count(self):
        """
        :returns: the number of undeleted documents in this set.
        """
        return sum(s.doc_count() for s in self.segments)
    
    
    def has_deletions(self):
        """
        :returns: True if this index has documents that are marked
            deleted but haven't been optimized out of the index yet.
            This includes deletions that haven't been written to disk
            with Index.commit() yet.
        """
        return any(s.has_deletions() for s in self.segments)
    
    def delete_document(self, docnum, delete = True):
        """Deletes a document by number.

        You must call Index.commit() for the deletion to be written to disk.
        """
        
        segment, segdocnum = self._segment_and_docnum(docnum)
        segment.delete_document(segdocnum, delete = delete)
    
    def deleted_count(self):
        """
        :returns: the total number of deleted documents in this index.
        """
        return sum(s.deleted_count() for s in self.segments)
    
    def is_deleted(self, docnum):
        """
        :returns: True if a given document number is deleted but not yet
            optimized out of the index.
        """
        
        segment, segdocnum = self._segment_and_docnum(docnum)
        return segment.is_deleted(segdocnum)
    
    def reader(self, storage, schema):
        from whoosh.filedb.filereading import SegmentReader
        segments = self.segments
        if len(segments) == 1:
            return SegmentReader(storage, segments[0], schema)
        else:
            from whoosh.reading import MultiReader
            readers = [SegmentReader(storage, segment, schema)
                       for segment in segments]
            return MultiReader(readers, self._doc_offsets, schema)
    

class Segment(object):
    """Do not instantiate this object directly. It is used by the Index
    object to hold information about a segment. A list of objects of this
    class are pickled as part of the TOC file.
    
    The TOC file stores a minimal amount of information -- mostly a list of
    Segment objects. Segments are the real reverse indexes. Having multiple
    segments allows quick incremental indexing: just create a new segment for
    the new documents, and have the index overlay the new segment over previous
    ones for purposes of reading/search. "Optimizing" the index combines the
    contents of existing segments into one (removing any deleted documents
    along the way).
    """
    
    def __init__(self, name, max_doc, field_length_totals, deleted = None):
        """
        :param name: The name of the segment (the Index object computes this from its
            name and the generation).
        :param max_doc: The maximum document number in the segment.
        :param term_count: Total count of all terms in all documents.
        :param field_length_totals: A dictionary mapping field numbers to the total
            number of terms in that field across all documents in the segment.
        :param deleted: A set of deleted document numbers, or None if no deleted
            documents exist in this segment.
        """
        
        self.name = name
        self.max_doc = max_doc
        self.field_length_totals = field_length_totals
        self.deleted = deleted
        
        self.doclen_filename = self.name + ".dci"
        self.docs_filename = self.name + ".dcz"
        self.term_filename = self.name + ".tiz"
        self.vector_filename = self.name + ".fvz"
        self.posts_filename = self.name + ".pst"
        self.vectorposts_filename = self.name + ".vps"
    
    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.name)
    
    def copy(self):
        if self.deleted:
            deleted = set(self.deleted)
        else:
            deleted = None
        return Segment(self.name, self.max_doc,
                       self.field_length_totals,
                       deleted)
    
    def doc_count_all(self):
        """
        :returns: the total number of documents, DELETED OR UNDELETED,
            in this segment.
        """
        return self.max_doc
    
    def doc_count(self):
        """:returns: the number of (undeleted) documents in this segment."""
        return self.max_doc - self.deleted_count()
    
    def has_deletions(self):
        """:returns: True if any documents in this segment are deleted."""
        return self.deleted_count() > 0
    
    def deleted_count(self):
        """:returns: the total number of deleted documents in this segment."""
        if self.deleted is None: return 0
        return len(self.deleted)
    
    def field_length(self, fieldnum):
        """
        :param fieldnum: the internal number of the field.
        :returns: the total number of terms in the given field across all
            documents in this segment.
        """
        return self.field_length_totals.get(fieldnum, 0)
    
    def delete_document(self, docnum, delete = True):
        """Deletes the given document number. The document is not actually
        removed from the index until it is optimized.

        :param docnum: The document number to delete.
        :param delete: If False, this undeletes a deleted document.
        """
        
        if delete:
            if self.deleted is None:
                self.deleted = set()
            elif docnum in self.deleted:
                raise KeyError("Document %s in segment %r is already deleted"
                               % (docnum, self.name))
            
            self.deleted.add(docnum)
        else:
            if self.deleted is None or docnum not in self.deleted:
                raise KeyError("Document %s is not deleted" % docnum)
            
            self.deleted.clear(docnum)
    
    def is_deleted(self, docnum):
        """:returns: True if the given document number is deleted."""
        
        if self.deleted is None: return False
        return docnum in self.deleted
    
   
# Utility functions

def _toc_pattern(indexname):
    """Returns a regular expression object that matches TOC filenames.
    name is the name of the index.
    """
    
    return re.compile("_%s_([0-9]+).toc" % indexname)

def _segment_pattern(indexname):
    """Returns a regular expression object that matches segment filenames.
    name is the name of the index.
    """
    
    return re.compile("(_%s_[0-9]+).(%s)" % (indexname, _EXTENSIONS))
            

























#===============================================================================
# Copyright 2009 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""
The classes in this module encode and decode posting information for a field.
The field format essentially determines what information is stored about each
occurance of a term.
"""

from collections import defaultdict
from struct import pack, unpack, calcsize
from cStringIO import StringIO

from whoosh.analysis import unstopped
from whoosh.system import _INT_SIZE, _USHORT_SIZE, _FLOAT_SIZE
from whoosh.util import varint, read_varint, float_to_byte, byte_to_float


# Format base class

class Format(object):
    """Abstract base class representing a storage format for a field or vector.
    Format objects are responsible for writing and reading the low-level
    representation of a field. It controls what kind/level of information
    to store about the indexed fields.
    """
    
    posting_size = -1
    textual = True
    __inittypes__ = dict(analyzer=object, field_boost=float)
    
    def __init__(self, analyzer, field_boost = 1.0, **options):
        """
        :param analyzer: The analysis.Analyzer object to use to index this field.
            See the analysis module for more information. If this value
            is None, the field is not indexed/searchable.
        :param field_boost: A constant boost factor to scale to the score
            of all queries matching terms in this field.
        """
        
        self.analyzer = analyzer
        self.field_boost = field_boost
        self.options = options
    
    def __eq__(self, other):
        return self.__class__ is other.__class__ and self.__dict__ == other.__dict__
    
    def __repr__(self):
        return "%s(%r, boost = %s)" % (self.__class__.__name__,
                                       self.analyzer, self.field_boost)
    
    def clean(self):
        if self.analyzer and hasattr(self.analyzer, "clean"):
            self.analyzer.clean()
    
    def word_values(self, value, **kwargs):
        """Takes the text value to be indexed and yields a series of
        ("tokentext", frequency, valuestring) tuples, where frequency is the
        number of times "tokentext" appeared in the value, and valuestring is
        encoded field-specific posting value for the token. For example, in a
        Frequency format, the value string would be the same as frequency; in
        a Positions format, the value string would encode a list of token
        positions at which "tokentext" occured.
        
        :param value: The unicode text to index.
        """
        raise NotImplementedError
    
    def analyze(self, unicodestring, mode='', **kwargs):
        """Returns a :class:`whoosh.analysis.Token` iterator from the given
        unicode string.
        
        :param unicodestring: the string to analyzer.
        :param mode: a string indicating the purpose for which the unicode
            string is being analyzed, i.e. 'index' or 'query'.
        """
        
        if not self.analyzer:
            raise Exception("%s format has no analyzer" % self.__class__)
        return self.analyzer(unicodestring, mode=mode, **kwargs)
    
    def encode(self, value):
        """Returns the given value encoded as a string.
        """
        raise NotImplementedError
    
    def supports(self, name):
        """Returns True if this format supports interpreting its posting
        value as 'name' (e.g. "frequency" or "positions").
        """
        return hasattr(self, "decode_" + name)
    
    def decoder(self, name):
        """Returns the bound method for interpreting value as 'name',
        where 'name' is for example "frequency" or "positions". This
        object must have a corresponding Format.decode_<name>() method.
        """
        return getattr(self, "decode_" + name)
    
    def decode_as(self, astype, valuestring):
        """Interprets the encoded value string as 'astype', where 'astype' is
        for example "frequency" or "positions". This object must have a
        corresponding decode_<astype>() method.
        """
        return self.decoder(astype)(valuestring)
    

# Concrete field classes

class Existence(Format):
    """Only indexes whether a given term occurred in a given document; it does
    not store frequencies or positions. This is useful for fields that should be
    searchable but not scorable, such as file path.
    
    Supports: frequency, weight (always reports frequency = 1).
    """
    
    posting_size = 0
    __inittypes__ = dict(analyzer=object, field_boost=float)
    
    def __init__(self, analyzer, field_boost = 1.0, **options):
        self.analyzer = analyzer
        self.field_boost = field_boost
        self.options = options
    
    def word_values(self, value, **kwargs):
        seen = set()
        for t in unstopped(self.analyzer(value, **kwargs)):
            seen.add(t.text)
        
        return ((w, 1, '') for w in seen)
    
    def encode(self, value):
        return ''
    
    def decode_frequency(self, valuestring):
        return 1
    
    def decode_weight(self, valuestring):
        return self.field_boost


class Frequency(Format):
    """Stores frequency information for each posting.
    
    Supports: frequency, weight.
    """
    
    posting_size = _INT_SIZE
    __inittypes__ = dict(analyzer=object, field_boost=float, boost_as_freq=bool)
    
    def __init__(self, analyzer, field_boost = 1.0, boost_as_freq = False, **options):
        """
        :param analyzer: The analysis.Analyzer object to use to index this field.
            See the analysis module for more information. If this value
            is None, the field is not indexed/searchable.
        :param field_boost: A constant boost factor to scale to the score
            of all queries matching terms in this field.
        :param boost_as_freq: if True, take the integer value of each token's
            boost attribute and use it as the token's frequency.
        """
        
        self.analyzer = analyzer
        self.field_boost = field_boost
        self.boost_as_freq = boost_as_freq
        self.options = options
        
    def word_values(self, value, **kwargs):
        seen = defaultdict(int)
        if self.boost_as_freq:
            for t in unstopped(self.analyzer(value, boosts = True, **kwargs)):
                seen[t.text] += int(t.boost)
        else:
            for t in unstopped(self.analyzer(value, **kwargs)):
                seen[t.text] += 1
        
        encode = self.encode
        return ((w, freq, encode(freq)) for w, freq in seen.iteritems())

    def encode(self, freq):
        return pack("!I", freq)
    
    def decode_frequency(self, valuestring):
        return unpack("!I", valuestring)[0]
    
    def decode_weight(self, valuestring):
        freq = unpack("!I", valuestring)[0]
        return freq * self.field_boost
    

class DocBoosts(Frequency):
    """A Field that stores frequency and per-document boost information for each posting.
    
    Supports: frequency, weight.
    """
    
    posting_size = _INT_SIZE + 1
    
    def word_values(self, value, doc_boost = 1.0, **kwargs):
        seen = defaultdict(int)
        for t in unstopped(self.analyzer(value, **kwargs)):
            seen[t.text] += 1
        
        encode = self.encode
        return ((w, freq, encode((freq, doc_boost))) for w, freq in seen.iteritems())
    
    def encode(self, freq_docboost):
        freq, docboost = freq_docboost
        return pack("!I", freq) + float_to_byte(docboost)
    
    def decode_docboosts(self, valuestring):
        freq = unpack("!I", valuestring[:_INT_SIZE])[0]
        docboost = byte_to_float(valuestring[-1])
        return (freq, docboost)
    
    def decode_frequency(self, valuestring):
        return unpack("!I", valuestring[0:_INT_SIZE])[0]
    
    def decode_weight(self, valuestring):
        freq = unpack("!I", valuestring[:_INT_SIZE])[0]
        docboost = byte_to_float(valuestring[-1])
        return freq * docboost * self.field_boost
    

# Vector formats

class Positions(Format):
    """A vector that stores position information in each posting, to allow phrase
    searching and "near" queries.
    
    Supports: frequency, weight, positions, position_boosts (always reports position
    boost = 1.0).
    """
    
    def word_values(self, value, start_pos = 0, **kwargs):
        seen = defaultdict(list)
        for t in unstopped(self.analyzer(value, positions = True, start_pos = start_pos, **kwargs)):
            seen[t.text].append(start_pos + t.pos)
        
        encode = self.encode
        return ((w, len(poslist), encode(poslist)) for w, poslist in seen.iteritems())
    
    def encode(self, positions):
        # positions = [pos1, pos2, ...]
        codes = []
        base = 0
        for pos in positions:
            codes.append(varint(pos - base))
            base = pos
        return pack("!I", len(positions)) + "".join(codes)
    
    def decode_positions(self, valuestring):
        read = StringIO(valuestring).read
        freq = unpack("!I", read(_INT_SIZE))[0]
        position = 0
        positions = []
        for _ in xrange(freq):
            position = read_varint(read) + position
            positions.append(position)
        return positions
    
    def decode_frequency(self, valuestring):
        return unpack("!I", valuestring[:_INT_SIZE])[0]
    
    def decode_weight(self, valuestring):
        return self.decode_frequency(valuestring) * self.field_boost
    
    def decode_position_boosts(self, valuestring):
        return [(pos, 1) for pos in self.decode_positions(valuestring)]
    

class Characters(Positions):
    """Stores token position and character start and end information
    for each posting.
    
    Supports: frequency, weight, positions, position_boosts (always reports position
    boost = 1.0), characters.
    """
    
    def word_values(self, value, start_pos = 0, start_char = 0, **kwargs):
        seen = defaultdict(list)
        
        for t in unstopped(self.analyzer(value, positions = True, chars = True,
                                         start_pos = start_pos, start_char = start_char, **kwargs)):
            seen[t.text].append((t.pos, start_char + t.startchar, start_char + t.endchar))
        
        encode = self.encode
        return ((w, len(ls), encode(ls)) for w, ls in seen.iteritems())
    
    def encode(self, posns_chars):
        # posns_chars = [(pos, startchar, endchar), ...]
        codes = []
        posbase = 0
        charbase = 0
        for pos, startchar, endchar in posns_chars:
            codes.append(varint(pos - posbase))
            posbase = pos
            codes.extend((varint(startchar - charbase), varint(endchar - startchar)))
            charbase = endchar
        return pack("!I", len(posns_chars)) + "".join(codes)
    
    def decode_characters(self, valuestring):
        read = StringIO(valuestring).read
        freq = unpack("!I", read(_INT_SIZE))[0]
        position = 0
        endchar = 0
        posns_chars = []
        for _ in xrange(freq):
            position = read_varint(read) + position
            startchar = endchar + read_varint(read)
            endchar = startchar + read_varint(read)
            posns_chars.append((position, startchar, endchar))
        return posns_chars
    
    def decode_positions(self, valuestring):
        return [pos for pos, startchar, endchar
                in self.decode_characters(valuestring)]
    

class PositionBoosts(Positions):
    """A format that stores positions and per-position boost information
    in each posting.
    
    Supports: frequency, weight, positions, position_boosts.
    """
    
    def word_values(self, value, start_pos = 0, **kwargs):
        seen = defaultdict(iter)
        for t in unstopped(self.analyzer(value, positions = True, boosts = True,
                                         start_pos = start_pos, **kwargs)):
            pos = t.pos
            boost = t.boost
            seen[t.text].append((pos, boost))
        
        encode = self.encode
        return ((w, len(poslist), encode(poslist)) for w, poslist in seen.iteritems())
    
    def encode(self, posns_boosts):
        # posns_boosts = [(pos, boost), ...]
        codes = []
        base = 0
        summedboost = 0
        for pos, boost in posns_boosts:
            summedboost += boost
            codes.extend((varint(pos - base), float_to_byte(boost)))
            base = pos
        
        return pack("!If", len(posns_boosts), summedboost) + "".join(codes)
    
    def decode_position_boosts(self, valuestring):
        f = StringIO(valuestring)
        read = f.read
        freq = unpack("!I", read(_INT_SIZE))[0]
        
        # Skip summed boost
        f.seek(_FLOAT_SIZE, 1)
        
        position = 0
        posns_boosts = []
        for _ in xrange(freq):
            position = read_varint(read) + position
            boost = byte_to_float(read(1))
            posns_boosts.append((position, boost))
        return posns_boosts
    
    def decode_positions(self, valuestring):
        f = StringIO(valuestring)
        read, seek = f.read, f.seek
        
        freq = unpack("!I", read(_INT_SIZE))[0]
        # Skip summed boost
        seek(_FLOAT_SIZE, 1)
        
        position = 0
        positions = []
        for _ in xrange(freq):
            position = read_varint(read) + position
            # Skip boost
            seek(1, 1)
            positions.append(position)
        return positions
    
    def decode_weight(self, valuestring):
        freq, summedboost = unpack("!If", valuestring[:_INT_SIZE + _FLOAT_SIZE])
        return freq * summedboost
    

class CharacterBoosts(Characters):
    """A format that stores positions, character start and end, and
    per-position boost information in each posting.
    
    Supports: frequency, weight, positions, position_boosts, characters,
    character_boosts.
    """
    
    def word_values(self, value, start_pos = 0, start_char = 0, **kwargs):
        seen = defaultdict(iter)
        for t in unstopped(self.analyzer(value, positions = True, characters = True,
                                         boosts = True,
                                         start_pos = start_pos, start_char = start_char,
                                         **kwargs)):
            seen[t.text].append((t.pos,
                                 start_char + t.startchar, start_char + t.endchar,
                                 t.boost))
        
        encode = self.encode
        return ((w, len(poslist), encode(poslist)) for w, poslist in seen.iteritems())
    
    def encode(self, posns_chars_boosts):
        # posns_chars_boosts = [(pos, startchar, endchar, boost), ...]
        codes = []
        
        posbase = 0
        charbase = 0
        summedboost = 0
        for pos, startchar, endchar, boost in posns_chars_boosts:
            summedboost += boost
            codes.append(varint(pos - posbase))
            posbase = pos
            codes.extend((varint(startchar - charbase),
                          varint(endchar - startchar),
                          float_to_byte(boost)))
            charbase = endchar
        
        return pack("!If", len(posns_chars_boosts), summedboost) + "".join(codes)
    
    def decode_character_boosts(self, valuestring):
        f = StringIO(valuestring)
        read = f.read
        
        freq = unpack("!I", read(_INT_SIZE))[0]
        # Skip summed boost
        f.seek(_FLOAT_SIZE, 1)
        
        position = 0
        endchar = 0
        posns_chars = []
        for _ in xrange(freq):
            position = read_varint(read) + position
            startchar = endchar + read_varint(read)
            endchar = startchar + read_varint(read)
            boost = byte_to_float(read(1))
            posns_chars.append((position, startchar, endchar, boost))
        return posns_chars
    
    def decode_characters(self, valuestring):
        return [(pos, startchar, endchar) for pos, startchar, endchar, boost
                in self.decode_character_boosts(valuestring)]
    
    def decode_position_boosts(self, valuestring):
        return [(pos, boost) for pos, startchar, endchar, boost
                in self.decode_character_boosts(valuestring)]
    








#===============================================================================
# Copyright 2007 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""Classes and functions for turning a piece of text into
an indexable stream of "tokens" (usually equivalent to words). There are
three general types of classes/functions involved in analysis:

* Tokenizers are always at the start of the text processing pipeline.
  They take a string and yield Token objects (actually, the same token
  object over and over, for performance reasons) corresponding to the
  tokens (words) in the text.
      
  Every tokenizer is a callable that takes a string and returns a
  generator of tokens.
      
* Filters take the tokens from the tokenizer and perform various
  transformations on them. For example, the LowercaseFilter converts
  all tokens to lowercase, which is usually necessary when indexing
  regular English text.
      
  Every filter is a callable that takes a token generator and returns
  a token generator.
      
* Analyzers are convenience functions/classes that "package up" a
  tokenizer and zero or more filters into a single unit, so you
  don't have to construct the tokenizer-filter-filter-etc. pipeline
  yourself. For example, the StandardAnalyzer combines a RegexTokenizer,
  LowercaseFilter, and StopFilter.
    
  Every analyzer is a callable that takes a string and returns a
  token generator. (So Tokenizers can be used as Analyzers if you
  don't need any filtering).
  
You can implement an analyzer as a custom class or function, or compose tokenizers
and filters together using the ``|`` character::

    my_analyzer = RegexTokenizer() | LowercaseFilter() | StopFilter()
    
The first item must be a tokenizer and the rest must be filters (you can't put a
filter first or a tokenizer after the first item).
"""

from array import array
import copy, re
from itertools import chain

from whoosh.lang.porter import stem

# Default list of stop words (words so common it's usually
# wasteful to index them). This list is used by the StopFilter
# class, which allows you to supply an optional list to override
# this one.

STOP_WORDS = frozenset(("the", "to", "of", "a", "and", "is", "in", "this",
                        "you", "for", "be", "on", "or", "will", "if", "can", "are",
                        "that", "by", "with", "it", "as", "from", "an", "when",
                        "not", "may", "tbd", "us", "we", "yet"))


# Utility functions

def unstopped(tokenstream):
    """Removes tokens from a token stream where token.stopped = True."""
    return (t for t in tokenstream if not t.stopped)


# Token object

class Token(object):
    """
    Represents a "token" (usually a word) extracted from the source text
    being indexed.
    
    See "Advanced analysis" in the user guide for more information.
    
    Because object instantiation in Python is slow, tokenizers should create
    ONE SINGLE Token object and YIELD IT OVER AND OVER, changing the attributes
    each time.
    
    This trick means that consumers of tokens (i.e. filters) must
    never try to hold onto the token object between loop iterations, or convert
    the token generator into a list.
    Instead, save the attributes between iterations, not the object::
    
        def RemoveDuplicatesFilter(self, stream):
            # Removes duplicate words.
            lasttext = None
            for token in stream:
                # Only yield the token if its text doesn't
                # match the previous token.
                if lasttext != token.text:
                    yield token
                lasttext = token.text

    """
    
    def __init__(self, positions = False, chars = False, boosts = False, removestops = True,
                 mode = '', **kwargs):
        """
        :param positions: Whether tokens should have the token position in
            the 'pos' attribute.
        :param chars: Whether tokens should have character offsets
            in the 'startchar' and 'endchar' attributes.
        :param boosts: whether the tokens should have per-token boosts
            in the 'boost' attribute.
        :param removestops: whether to remove stop words from the stream
            (if the tokens pass through a stop filter).
        :param mode: contains a string describing the purpose for which the
            analyzer is being called, i.e. 'index' or 'query'.
        """
        
        self.positions = positions
        self.chars = chars
        self.boosts = boosts
        self.stopped = False
        self.boost = 1.0
        self.removestops = removestops
        self.mode = mode
        self.__dict__.update(kwargs)
    
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__,
                           ", ".join(["%s=%r" % (name, value)
                                      for name, value in self.__dict__.iteritems()]))
        
    def copy(self):
        return copy.copy(self)


# Composition support

class Composable(object):
    def __or__(self, other):
        assert callable(other), "%r is not callable" % other
        return CompositeAnalyzer(self, other)
    
    def __repr__(self):
        attrs = ""
        if self.__dict__:
            attrs = ", ".join("%s=%r" % (key, value) for key, value in self.__dict__.iteritems())
        return self.__class__.__name__ + "(%s)" % attrs


# Tokenizers

class Tokenizer(Composable):
    """Base class for Tokenizers.
    """
    
    def __eq__(self, other):
        return other and self.__class__ is other.__class__


class IDTokenizer(Tokenizer):
    """Yields the entire input string as a single token. For use
    in indexed but untokenized fields, such as a document's path.
    
    >>> idt = IDTokenizer()
    >>> [token.text for token in idt(u"/a/b 123 alpha")]
    [u"/a/b 123 alpha"]
    """
    
    def __call__(self, value, positions = False, chars = False,
                 keeporiginal = False, removestops = True,
                 start_pos = 0, start_char = 0, mode='',
                 **kwargs):
        assert isinstance(value, unicode), "%r is not unicode" % value
        t = Token(positions, chars, removestops = removestops, mode=mode)
        t.text = value
        if keeporiginal:
            t.original = value
        if positions:
            t.pos = start_pos + 1
        if chars:
            t.startchar = start_char
            t.endchar = start_char + len(value)
        yield t
    

class RegexTokenizer(Tokenizer):
    """
    Uses a regular expression to extract tokens from text.
    
    >>> rex = RegexTokenizer()
    >>> [token.text for token in rex(u"hi there 3.141 big-time under_score")]
    [u"hi", u"there", u"3.141", u"big", u"time", u"under_score"]
    """
    
    __inittypes__ = dict(expression=unicode, gaps=bool)
    
    def __init__(self, expression = r"\w+(\.?\w+)*", gaps=False):
        """
        :param expression: A regular expression object or string. Each match
            of the expression equals a token. Group 0 (the entire matched text)
            is used as the text of the token. If you require more complicated
            handling of the expression match, simply write your own tokenizer.
        :param gaps: If True, the tokenizer *splits* on the expression, rather
            than matching on the expression.
        """
        
        if isinstance(expression, basestring):
            self.expression = re.compile(expression, re.UNICODE)
        else:
            self.expression = expression
        self.gaps = gaps
    
    def __eq__(self, other):
        if self.__class__ is other.__class__:
            if self.expression.pattern == other.expression.pattern:
                return True
        return False
    
    def __call__(self, value, positions = False, chars = False,
                 keeporiginal = False, removestops = True,
                 start_pos = 0, start_char = 0,
                 tokenize = True, mode = '', **kwargs):
        """
        :param value: The unicode string to tokenize.
        :param positions: Whether to record token positions in the token.
        :param chars: Whether to record character offsets in the token.
        :param start_pos: The position number of the first token. For example,
            if you set start_pos=2, the tokens will be numbered 2,3,4,...
            instead of 0,1,2,...
        :param start_char: The offset of the first character of the first
            token. For example, if you set start_char=2, the text "aaa bbb"
            will have chars (2,5),(6,9) instead (0,3),(4,7).
        :param tokenize: if True, the text should be tokenized. 
        """
        
        assert isinstance(value, unicode), "%r is not unicode" % value
        
        t = Token(positions, chars, removestops=removestops, mode=mode)
        if not tokenize:
            t.original = t.text = value
            if positions: t.pos = start_pos
            if chars:
                t.startchar = start_char
                t.endchar = start_char + len(value)
            yield t
        elif not self.gaps:
            # The default: expression matches are used as tokens
            for pos, match in enumerate(self.expression.finditer(value)):
                t.text = match.group(0)
                if keeporiginal:
                    t.original = t.text
                t.stopped = False
                if positions:
                    t.pos = start_pos + pos
                if chars:
                    t.startchar = start_char + match.start()
                    t.endchar = start_char + match.end()
                yield t
        else:
            # When gaps=True, iterate through the matches and
            # yield the text between them.
            prevend = 0
            pos = start_pos
            for match in self.expression.finditer(value):
                start = prevend
                end = match.start()
                text = value[start:end]
                if text:
                    t.text = text
                    if keeporiginal:
                        t.original = t.text
                    t.stopped = False
                    if positions:
                        t.pos = pos
                        pos += 1
                    if chars:
                        t.startchar = start_char + start
                        t.endchar = start_char + end
                    
                    yield t
                
                prevend = match.end()
            
            # If the last "gap" was before the end of the text,
            # yield the last bit of text as a final token.
            if prevend < len(value):
                t.text = value[prevend:]
                if keeporiginal:
                    t.original = t.text
                t.stopped = False
                if positions:
                    t.pos = pos
                if chars:
                    t.startchar = prevend
                    t.endchar = len(value)
                yield t


class CharsetTokenizer(Tokenizer):
    """Tokenizes and translates text according to a character mapping object. Characters
    that map to None are considered token break characters. For all other characters the
    map is used to translate the character. This is useful for case and accent folding.
    
    This tokenizer loops character-by-character and so will likely be much
    slower than :class:`RegexTokenizer`.
    
    One way to get a character mapping object is to convert a Sphinx charset table file
    using :func:`whoosh.support.charset.charset_table_to_dict`.
    
    >>> from whoosh.support.charset import charset_table_to_dict, default_charset
    >>> charmap = charset_table_to_dict(default_charset)
    >>> chtokenizer = CharsetTokenizer(charmap)
    >>> [t.text for t in chtokenizer(u'Stra\\xdfe ABC')]
    [u'strase', u'abc']
    
    The Sphinx charset table format is described at
    http://www.sphinxsearch.com/docs/current.html#conf-charset-table.
    """
    
    __inittype__ = dict(charmap=str)
    
    def __init__(self, charmap):
        """
        :param charmap: a mapping from integer character numbers to unicode characters,
            as used by the unicode.translate() method.
        """
        self.charmap = charmap
    
    def __eq__(self, other):
        return other and self.__class__ is other.__class__ and self.charmap == other.charmap

    def __call__(self, value, positions = False, chars = False,
             keeporiginal = False, removestops = True,
             start_pos = 0, start_char = 0,
             tokenize = True, mode = '', **kwargs):
        """
        :param value: The unicode string to tokenize.
        :param positions: Whether to record token positions in the token.
        :param chars: Whether to record character offsets in the token.
        :param start_pos: The position number of the first token. For example,
            if you set start_pos=2, the tokens will be numbered 2,3,4,...
            instead of 0,1,2,...
        :param start_char: The offset of the first character of the first
            token. For example, if you set start_char=2, the text "aaa bbb"
            will have chars (2,5),(6,9) instead (0,3),(4,7).
        :param tokenize: if True, the text should be tokenized. 
        """
        
        assert isinstance(value, unicode), "%r is not unicode" % value
        
        t = Token(positions, chars, removestops=removestops, mode=mode)
        if not tokenize:
            t.original = t.text = value
            if positions: t.pos = start_pos
            if chars:
                t.startchar = start_char
                t.endchar = start_char + len(value)
            yield t
        else:
            text = u""
            charmap = self.charmap
            pos = start_pos
            startchar = currentchar = start_char
            for char in value:
                tchar = charmap[ord(char)]
                if tchar:
                    text += tchar
                else:
                    if currentchar > startchar:
                        t.text = text
                        if keeporiginal:
                            t.original = t.text
                        if positions:
                            t.pos = pos
                            pos += 1
                        if chars:
                            t.startchar = startchar
                            t.endchar = currentchar
                        yield t
                    startchar = currentchar + 1
                    text = u""
                    
                currentchar += 1
            
            if currentchar > startchar:
                t.text = value[startchar:currentchar]
                if keeporiginal:
                    t.original = t.text
                if positions:
                    t.pos = pos
                if chars:
                    t.startchar = startchar
                    t.endchar = currentchar
                yield t


def SpaceSeparatedTokenizer():
    """Returns a RegexTokenizer that splits tokens by whitespace.
    
    >>> sst = SpaceSeparatedTokenizer()
    >>> [token.text for token in sst(u"hi there big-time, what's up")]
    [u"hi", u"there", u"big-time,", u"what's", u"up"]
    """
    
    return RegexTokenizer(r"[^ \t\r\n]+")


def CommaSeparatedTokenizer():
    """Splits tokens by commas.
    
    Note that the tokenizer calls unicode.strip() on each match
    of the regular expression.
    
    >>> cst = CommaSeparatedTokenizer()
    >>> [token.text for token in cst(u"hi there, what's , up")]
    [u"hi there", u"what's", u"up"]
    """
    
    return RegexTokenizer(r"[^,]+") | StripFilter()


class NgramTokenizer(Tokenizer):
    """Splits input text into N-grams instead of words.
    
    >>> ngt = NgramTokenizer(4)
    >>> [token.text for token in ngt(u"hi there")]
    [u"hi t", u"i th", u" the", u"ther", u"here"]
    
    Note that this tokenizer does NOT use a regular expression to extract words,
    so the grams emitted by it will contain whitespace, punctuation, etc. You may
    want to massage the input or add a custom filter to this tokenizer's output.
    
    Alternatively, if you only want sub-word grams without whitespace, you
    could combine a RegexTokenizer with NgramFilter instead.
    """
    
    __inittypes__ = dict(minsize=int, maxsize=int)
    
    def __init__(self, minsize, maxsize = None):
        """
        :param minsize: The minimum size of the N-grams.
        :param maxsize: The maximum size of the N-grams. If you omit
            this parameter, maxsize == minsize.
        """
        
        self.min = minsize
        self.max = maxsize or minsize
    
    def __eq__(self, other):
        if self.__class__ is other.__class__:
            if self.min == other.min and self.max == other.max:
                return True
        return False
    
    def __call__(self, value, positions = False, chars = False,
                 keeporiginal = False, removestops = True,
                 start_pos = 0, start_char = 0,
                 **kwargs):
        assert isinstance(value, unicode), "%r is not unicode" % value
        
        inlen = len(value)
        t = Token(positions, chars, removestops = removestops)
        pos = start_pos
        for start in xrange(0, inlen - self.min + 1):
            for size in xrange(self.min, self.max + 1):
                end = start + size
                if end > inlen: continue
                
                t.text = value[start:end]
                if keeporiginal:
                    t.original = t.text
                t.stopped = False
                if positions:
                    t.pos = pos
                if chars:
                    t.startchar = start_char + start
                    t.endchar = start_char + end
                
                yield t
            pos += 1
                    

# Filters

class Filter(Composable):
    """Base class for Filter objects. A Filter subclass must implement
    a __call__ method that takes a single argument, which is an iterator
    of Token objects, and yield a series of Token objects in return.
    """
    
    def __eq__(self, other):
        return other and self.__class__ is other.__class__


class PassFilter(Filter):
    """An identity filter: passes the tokens through untouched.
    """
    
    def __call__(self, tokens):
        assert hasattr(tokens, "__iter__")
        for t in tokens:
            yield t


class RecordFilter(Filter):
    """A debug filter that remembers the tokens that pass through
    it, and stores them in the 'tokens' attribute.
    """
    
    def __init__(self):
        self.tokens = None
    
    def __call__(self, tokens):
        assert hasattr(tokens, "__iter__")
        self.tokens = []
        for t in tokens:
            self.tokens.append(t.copy())
            yield t


class MultiFilter(Filter):
    """Chooses one of two or more sub-filters based on the 'mode' attribute
    of the token stream.
    """
    
    def __init__(self, **kwargs):
        """Use keyword arguments to associate mode attribute values with
        instantiated filters.
        
        >>> iwf_for_index = IntraWordFilter(mergewords=True, mergenums=False)
        >>> iwf_for_query = IntraWordFilter(mergewords=False, mergenums=False)
        >>> mf = MultiFilter(index=iwf_for_index, query=iwf_for_query)
        
        This class expects that the value of the mode attribute is consistent
        among all tokens in a token stream.
        """
        self.filters = kwargs
    
    def __eq__(self, other):
        return other and self.__class__ is other.__class__ and self.filters == other.filters
    
    def __call__(self, tokens):
        # Only selects on the first token
        t = tokens.next()
        filter = self.filters[t.mode]
        return filter(chain([t], tokens))
        

class LowercaseFilter(Filter):
    """Uses unicode.lower() to lowercase token text.
    
    >>> rext = RegexTokenizer()
    >>> stream = rext(u"This is a TEST")
    >>> [token.text for token in LowercaseFilter(stream)]
    [u"this", u"is", u"a", u"test"]
    """
    
    def __call__(self, tokens):
        assert hasattr(tokens, "__iter__")
        for t in tokens:
            t.text = t.text.lower()
            yield t
            

class StripFilter(Filter):
    """Calls unicode.strip() on the token text.
    
    >>> rext = CommaSeparatedTokenizer()
    >>> stream = rext(u"This i
    """
    
    def __call__(self, tokens):
        assert hasattr(tokens, "__iter__")
        for t in tokens:
            t.text = t.text.strip()
            yield t


class StopFilter(Filter):
    """Marks "stop" words (words too common to index) in the stream (and by default
    removes them).
    
    >>> rext = RegexTokenizer()
    >>> stream = rext(u"this is a test")
    >>> stopper = StopFilter()
    >>> [token.text for token in sopper(stream)]
    [u"this", u"test"]
    
    """

    __inittypes__ = dict(stoplist=list, minsize=int, renumber=bool)

    def __init__(self, stoplist = STOP_WORDS, minsize = 2,
                 renumber = True):
        """
        :param stoplist: A collection of words to remove from the stream.
            This is converted to a frozenset. The default is a list of
            common stop words.
        :param minsize: The minimum length of token texts. Tokens with
            text smaller than this will be stopped.
        :param renumber: Change the 'pos' attribute of unstopped tokens
            to reflect their position with the stopped words removed.
        :param remove: Whether to remove the stopped words from the stream
            entirely. This is not normally necessary, since the indexing
            code will ignore tokens it receives with stopped=True.
        """
        
        if stoplist is None:
            self.stops = frozenset()
        else:
            self.stops = frozenset(stoplist)
        self.min = minsize
        self.renumber = renumber
    
    def __eq__(self, other):
        if self.__class__ is other.__class__:
            if self.stops == other.stops and self.min == other.min and self.renumber == other.renumber:
                return True
        return False
    
    def __call__(self, tokens):
        assert hasattr(tokens, "__iter__")
        stoplist = self.stops
        minsize = self.min
        renumber = self.renumber
        
        pos = None
        for t in tokens:
            text = t.text
            if len(text) >= minsize and text not in stoplist:
                # This is not a stop word
                if renumber and t.positions:
                    if pos is None:
                        pos = t.pos
                    else:
                        pos += 1
                    t.pos = pos
                t.stopped = False
                yield t
            else:
                # This is a stop word
                if not t.removestops:
                    # This IS a stop word, but we're not removing them
                    t.stopped = True
                    yield t


class StemFilter(Filter):
    """Stems (removes suffixes from) the text of tokens using the Porter stemming
    algorithm. Stemming attempts to reduce multiple forms of the same root word
    (for example, "rendering", "renders", "rendered", etc.) to a single word in
    the index.
    
    >>> rext = RegexTokenizer()
    >>> stream = rext(u"fundamentally willows")
    >>> stemmer = StemFilter()
    >>> [token.text for token in stemmer(stream)]
    [u"fundament", u"willow"]
    """
    
    __inittypes__ = dict(stemfn=object, ignore=list)
    
    def __init__(self, stemfn = stem, ignore = None):
        """
        :param stemfn: the function to use for stemming.
        :param ignore: a set/list of words that should not be stemmed. This
            is converted into a frozenset. If you omit this argument, all tokens
            are stemmed.
        """
        
        self.stemfn = stemfn
        self.cache = {}
        if ignore is None:
            self.ignores = frozenset()
        else:
            self.ignores = frozenset(ignore)
    
    def __eq__(self, other):
        return other and self.__class__ is other.__class__ and self.stemfn == other.stemfn
    
    def __call__(self, tokens):
        assert hasattr(tokens, "__iter__")
        stemfn = self.stemfn
        cache = self.cache
        ignores = self.ignores
        
        for t in tokens:
            if t.stopped:
                yield t
                continue
            
            text = t.text
            if text in ignores:
                yield t
            elif text in cache:
                t.text = cache[text]
                yield t
            else:
                t.text = s = stemfn(text)
                cache[text] = s
                yield t
                
    def clean(self):
        """
        This filter memoizes previously stemmed words to greatly speed up
        stemming. This method clears the cache of previously stemmed words.
        """
        self.cache.clear()


class CharsetFilter(Filter):
    """Translates the text of tokens by calling unicode.translate() using the supplied
    character mapping object. This is useful for case and accent folding.
    
    One way to get a character mapping object is to convert a Sphinx charset table file
    using :func:`whoosh.support.charset.charset_table_to_dict`.
    
    >>> from whoosh.support.charset import charset_table_to_dict, default_charset
    >>> retokenizer = RegexTokenizer()
    >>> charmap = charset_table_to_dict(default_charset)
    >>> chfilter = CharsetFilter(charmap)
    >>> [t.text for t in chfilter(retokenizer(u'Stra\\xdfe'))]
    [u'strase']
    
    The Sphinx charset table format is described at
    http://www.sphinxsearch.com/docs/current.html#conf-charset-table.
    """
    
    __inittypes__ = dict(charmap=str)
    
    def __init__(self, charmap):
        """
        :param charmap: a mapping from integer character numbers to unicode characters,
            as required by the unicode.translate() method.
        """
        self.charmap = charmap
    
    def __eq__(self, other):
        return other and self.__class__ is other.__class__ and self.charmap == other.charmap
    
    def __call__(self, tokens):
        assert hasattr(tokens, "__iter__")
        charmap = self.charmap
        for t in tokens:
            t.text = t.text.translate(charmap)
            yield t


class NgramFilter(Filter):
    """Splits token text into N-grams.
    
    >>> rext = RegexTokenizer()
    >>> stream = rext(u"hello there")
    >>> ngf = NgramFilter(4)
    >>> [token.text for token in ngf(stream)]
    [u"hell", u"ello", u"ther", u"here"]
    
    """
    
    __inittypes__ = dict(minsize=int, maxsize=int)
    
    def __init__(self, minsize, maxsize = None):
        """
        :param minsize: The minimum size of the N-grams.
        :param maxsize: The maximum size of the N-grams. If you omit
            this parameter, maxsize == minsize.
        """
        
        self.min = minsize
        self.max = maxsize or minsize
    
    def __eq__(self, other):
        return other and self.__class__ is other.__class__\
        and self.min == other.min and self.max == other.max
    
    def __call__(self, tokens):
        assert hasattr(tokens, "__iter__")
        for t in tokens:
            text, chars = t.text, t.chars
            if chars:
                startchar = t.startchar
            # Token positions don't mean much for N-grams,
            # so we'll leave the token's original position
            # untouched.
            
            for start in xrange(0, len(text) - self.min):
                for size in xrange(self.min, self.max + 1):
                    end = start + size
                    if end > len(text): continue
                    
                    t.text = text[start:end]
                    
                    if chars:
                        t.startchar = startchar + start
                        t.endchar = startchar + end
                        
                    yield t


class IntraWordFilter(Filter):
    """Splits words into subwords and performs optional transformations on subword groups.
    This filter funtionally based on yonik's WordDelimiterFilter in Solr, but shares no
    code with it.
    
    * Split on intra-word delimiters, e.g. `Wi-Fi` -> `Wi`, `Fi`.
    * When splitwords=True, split on case transitions, e.g. `PowerShot` -> `Power`, `Shot`.
    * When splitnums=True, split on letter-number transitions, e.g. `SD500` -> `SD`, `500`.
    * Leading and trailing delimiter characters are ignored.
    * Trailing possesive "'s" removed from subwords, e.g. `O'Neil's` -> `O`, `Neil`.
    
    The mergewords and mergenums arguments turn on merging of subwords.
    
    When the merge arguments are false, subwords are not merged.
    
    * `PowerShot` -> `0`:`Power`, `1`:`Shot` (where `0` and `1` are token positions).
    
    When one or both of the merge arguments are true, consecutive runs of alphabetic
    and/or numeric subwords are merged into a new token with the same token position
    as the last sub-word.
    
    * `PowerShot` -> `0`:`Power`, `1`:`Shot`, `1`:`PowerShot`
    * `A's+B's&C's` -> `0`:`A`, `1`:`B`, `2`:`C`, `2`:`ABC`
    * `Super-Duper-XL500-42-AutoCoder!` -> 0:`Super`, 1:`Duper`, 2:`XL`, 2:`SuperDuperXL`,
      `3`:`500`, `4`:`42`, `4`:`50042`, `5`:`Auto`, `6`:`Coder`, `6`:`AutoCoder`
    
    When using this filter you should use a tokenizer that only splits on whitespace,
    so the tokenizer does not remove intra-word delimiters before this filter can see them,
    and put this filter before any use of LowercaseFilter.
    
    >>> analyzer = RegexTokenizer(r"\\S+") | IntraWordFilter() | LowercaseFilter()
    
    One use for this filter is to help match different written representations of a
    concept. For example, if the source text contained `wi-fi`, you probably want
    `wifi`, `WiFi`, `wi-fi`, etc. to match. One way of doing this is to specify
    mergewords=True and/or mergenums=True in the analyzer used for indexing, and
    mergewords=False / mergenums=False in the analyzer used for querying.
    
    >>> iwf = MultiFilter(index=IntraWordFilter(mergewords=True, mergenums=True),
                          query=IntraWordFilter(mergewords=False, mergenums=False))
    >>> analyzer = RegexTokenizer(r"\\S+") | iwf | LowercaseFilter()
    
    (See :class:`MultiFilter`.)
    """

    # Create sets of unicode digit, uppercase, and lowercase characters.
    digits = array("u")
    uppers = array("u")
    lowers = array("u")
    for n in xrange(2**16-1):
        ch = unichr(n)
        if ch.islower(): lowers.append(ch)
        elif ch.isupper(): uppers.append(ch)
        elif ch.isdigit(): digits.append(ch)
    
    # Create escaped strings of characters for use in regular expressions
    digits = re.escape("".join(digits))
    uppers = re.escape("".join(uppers))
    lowers = re.escape("".join(lowers))
    letters = uppers + lowers
    
    __inittypes__ = dict(delims=unicode, splitwords=bool, splitnums=bool,
                         mergewords=bool, mergenums=bool)
    
    def __init__(self, delims=u"-_'\"()!@#$%^&*[]{}<>\|;:,./?`~=+",
                 splitwords=True, splitnums=True,
                 mergewords=False, mergenums=False):
        """
        :param delims: a string of delimiter characters.
        :param splitwords: split at case transitions, e.g. `PowerShot` -> `Power`, `Shot`
        :param splitnums: split at letter-number transitions, e.g. `SD500` -> `SD`, `500`
        :param mergewords: merge consecutive runs of alphabetic subwords into an
            additional token with the same position as the last subword.
        :param mergenums: merge consecutive runs of numeric subwords into an
            additional token with the same position as the last subword.
        """
        
        self.delims = re.escape(delims)
        
        # Expression for splitting at delimiter characters
        self.splitter = re.compile(u"[%s]+" % (self.delims, ), re.UNICODE)
        # Expression for removing "'s" from the end of sub-words
        self.disposses = re.compile(u"(?<=[%s])'[Ss](?=$|[%s])" % (self.letters,
                                                                   self.delims), re.UNICODE)
        
        # Expression for finding case and letter-number transitions
        lower2upper = u"[%s][%s]" % (self.lowers, self.uppers)
        letter2digit = u"[%s][%s]" % (self.letters, self.digits)
        digit2letter = u"[%s][%s]" % (self.digits, self.letters)
        if splitwords and splitnums:
            self.boundary = re.compile(u"(%s|%s|%s)" % (lower2upper,
                                                        letter2digit,
                                                        digit2letter), re.UNICODE)
        elif splitwords:
            self.boundary = re.compile(u"%s" % (lower2upper, ), re.UNICODE)
        elif splitnums:
            self.boundary = re.compile(u"(%s|%s)" % (letter2digit,
                                                     digit2letter), re.UNICODE)
        
        self.splitting = splitwords or splitnums
        self.mergewords = mergewords
        self.mergenums = mergenums
    
    def __eq__(self, other):
        return other and self.__class__ is other.__class__\
        and self.__dict__ == other.__dict__
    
    def split(self, string):
        boundaries = self.boundary.finditer
        
        # Are we splitting on word/num boundaries?
        if self.splitting:
            parts = []
            # First, split on delimiters
            splitted = self.splitter.split(string)
            
            for run in splitted:
                # For each delimited run of characters, find the
                # boundaries (e.g. lower->upper, letter->num, num->letter)
                # and split between them.
                start = 0
                for match in boundaries(run):
                    middle = match.start() + 1
                    parts.append(run[start:middle])
                    start = middle
                    
                # Add the bit after the last split
                if start < len(run):
                    parts.append(run[start:])
        else:
            # Just split on delimiters
            parts = self.splitter.split(string)
        return parts
    
    def merge(self, parts):
        mergewords = self.mergewords
        mergenums = self.mergenums
        
        # Current type (1=alpah, 2=digit)
        last = 0
        # Where to insert a merged term in the original list
        insertat = 0
        # Buffer for parts to merge
        buf = []
        for pos, part in parts[:]:
            # Set the type of this part
            if part.isalpha(): this = 1
            elif part.isdigit(): this = 2
            
            # Is this the same type as the previous part?
            if buf and (this == last == 1 and mergewords)\
            or (this == last == 2 and mergenums):
                # This part is the same type as the previous.
                # Add it to the buffer of parts to merge.
                buf.append(part)
            else:
                # This part is different than the previous.
                if len(buf) > 1:
                    # If the buffer has at least two parts in
                    # it, merge them and add them to the original
                    # list of parts.
                    parts.insert(insertat, (pos-1, u"".join(buf)))
                    insertat += 1
                # Reset the buffer
                buf = [part]
                last = this
            insertat += 1
        
        # If there are parts left in the buffer at the end,
        # merge them and add them to the original list.
        if len(buf) > 1:
            parts.append((pos, u"".join(buf)))
    
    def __call__(self, tokens):
        disposses = self.disposses.sub
        merge = self.merge
        mergewords = self.mergewords
        mergenums = self.mergenums
        
        # This filter renumbers tokens as it expands them.
        # New position counter.
        
        newpos = None
        for t in tokens:
            text = t.text
            
            # If this is the first token we've seen, use it to set
            # the new position counter
            if newpos is None:
                if t.positions:
                    newpos = t.pos
                else:
                    # Token doesn't have positions, just use 0
                    newpos = 0
            
            if (text.isalpha() and (text.islower() or text.isupper())) or text.isdigit():
                # Short-circuit the common cases of no delimiters, no case transitions,
                # only digits, etc.
                t.pos = newpos
                yield t
                newpos += 1
            else:
                # Should we check for an apos before doing the
                # disposses step? Or is the re faster?
                # if "'" in text:
                text = disposses("", text)
                
                # Split the token text on delimiters, word and/or number boundaries,
                # and give the split parts positions
                parts = [(newpos + i, part) for i, part in enumerate(self.split(text))]
                
                # Did the split yield more than one part?
                if len(parts) > 1:
                    # If the options are set, merge consecutive runs of all-letters
                    # and/or all-numbers.
                    if mergewords or mergenums:
                        merge(parts)
                    
                    # Yield tokens for the parts
                    for pos, text in parts:
                        t.text = text
                        t.pos = pos
                        yield t
                    
                    # Set the new position counter based on the last part
                    newpos = parts[-1][0] + 1
                else:
                    # The split only gave one part, so just yield the 
                    # "dispossed" text.
                    t.text = text
                    t.pos = newpos
                    yield t
                    newpos += 1


class CamelFilter(Filter):
    """Splits CamelCased words into multiple words.
    This filter is deprecated, use IntraWordFilter instead.
    
    >>> rext = RegexTokenizer()
    >>> stream = rext(u"call getProcessedToken")
    >>> [token.text for token in CamelFilter(stream)]
    [u"call", u"getProcessedToken", u"get", u"Processed", u"Token"]
    
    Obviously this filter needs to precede LowercaseFilter if they
    are both in a filter chain.
    """
    
    camel_exp = re.compile("[A-Z][a-z]*|[a-z]+|[0-9]+")
    
    def __call__(self, tokens):
        assert hasattr(tokens, "__iter__")
        camel_exp = self.camel_exp
        for t in tokens:
            yield t
            text = t.text
            
            if text and not text.islower() and not text.isupper() and not text.isdigit():
                chars = t.chars
                if chars:
                    oldstart = t.startchar
                
                for match in camel_exp.finditer(text):
                    sub = match.group(0)
                    if sub != text:
                        t.text = sub
                        if chars:
                            t.startchar = oldstart + match.start()
                            t.endchar = oldstart + match.end()
                        yield t


class UnderscoreFilter(Filter):
    """Splits words with underscores into multiple words.
    This filter is deprecated, use IntraWordFilter instead.
    
    >>> rext = RegexTokenizer()
    >>> stream = rext(u"call get_processed_token")
    >>> [token.text for token in CamelFilter(stream)]
    [u"call", u"get_processed_token", u"get", u"processed", u"token"]
    
    Obviously you should not split words on underscores in the
    tokenizer if you want to use this filter.
    """
    
    underscore_exp = re.compile("[A-Z][a-z]*|[a-z]+|[0-9]+")
    
    def __call__(self, tokens):
        underscore_exp = self.underscore_exp
        for t in tokens:
            yield t
            text = t.text
            
            if text:
                chars = t.chars
                if chars:
                    oldstart = t.startchar
                
                for match in underscore_exp.finditer(text):
                    sub = match.group(0)
                    if sub != text:
                        t.text = sub
                        if chars:
                            t.startchar = oldstart + match.start()
                            t.endchar = oldstart + match.end()
                        yield t


class BoostTextFilter(Filter):
    """Advanced filter. Looks for embedded boost markers in the actual text of
    each token and extracts them to set the token's boost. This might be useful
    to let users boost individual terms.
    
    For example, if you added a filter:
    
      BoostTextFilter("\\^([0-9.]+)$")
    
    The user could then write keywords with an optional boost encoded in them,
    like this:
    
      image render^2 file^0.5
    
    (Of course, you might want to write a better pattern for the number part.)
    
     - Note that the pattern is run on EACH TOKEN, not the source text as a whole.
     
     - Because this filter runs a regular expression match on every token,
       for performance reasons it is probably only suitable for short fields.
       
     - You may use this filter in a Frequency-formatted field, where
       the Frequency format object has boost_as_freq = True. Bear in mind that
       in that case, you can only use integer "boosts".
    """
    
    def __init__(self, expression, group = 1, default = 1.0):
        """
        :param expression: a compiled regular expression object representing
        the pattern to look for within each token.
        :param group: the group name or number to use as the boost number
            (what to pass to match.group()). The string value of this group is
            passed to float().
        :param default: the default boost to use for tokens that don't have
            the marker.
        """
        
        self.expression = expression
        self.group = group
        self.default = default
    
    def __eq__(self, other):
        return other and self.__class__ is other.__class__\
        and self.expression == other.expression and self.default == other.default\
        and self.group == other.group
    
    def __call__(self, tokens):
        expression = self.expression
        groupnum = self.group
        default = self.default
    
        for t in tokens:
            text = t.text
            m = expression.match(text)
            if m:
                text = text[:m.start()] + text[m.end():]
                t.boost = float(m.group(groupnum))
            else:
                t.boost = default
                
            yield t

# Analyzers

class Analyzer(Composable):
    """
    Abstract base class for analyzers. Since the analyzer protocol is just
    __call__, this is pretty simple -- it mostly exists to provide common
    implementations of __repr__ and __eq__.
    """
    
    def __repr__(self):
        return "%s()" % self.__class__.__name__

    def __eq__(self, other):
        return other and self.__class__ is other.__class__ and self.__dict__ == other.__dict__

    def __call__(self, value, **kwargs):
        raise NotImplementedError
    
    def clean(self):
        pass


class CompositeAnalyzer(Analyzer):
    def __init__(self, *composables):
        self.items = []
        for comp in composables:
            if isinstance(comp, CompositeAnalyzer):
                self.items.extend(comp.items)
            else:
                self.items.append(comp)
    
    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__,
                           ", ".join(repr(item) for item in self.items))
    
    def __call__(self, value, **kwargs):
        items = self.items
        gen = items[0](value, **kwargs)
        for item in items[1:]:
            gen = item(gen)
        return gen
    
    def __getitem__(self, item):
        return self.items.__getitem__(item)
    
    def __len__(self):
        return len(self.items)
    
    def __eq__(self, other):
        return other and self.__class__ is other.__class__ and self.items == other.items
    
    def clean(self):
        for item in self.items:
            if hasattr(item, "clean"):
                item.clean()


def IDAnalyzer(lowercase = False):
    """Deprecated, just use an IDTokenizer directly, with a LowercaseFilter if desired.
    """
    
    tokenizer = IDTokenizer()
    if lowercase:
        tokenizer = tokenizer | LowercaseFilter()
    return tokenizer
IDAnalyzer.__inittypes__ = dict(lowercase=bool)


def KeywordAnalyzer(lowercase = False, commas = False):
    """Parses space-separated tokens.
    
    >>> ana = KeywordAnalyzer()
    >>> [token.text for token in ana(u"Hello there, this is a TEST")]
    [u"Hello", u"there,", u"this", u"is", u"a", u"TEST"]
    
    :param lowercase: whether to lowercase the tokens.
    :param commas: if True, items are separated by commas rather than spaces.
    """
    
    if commas:
        tokenizer = CommaSeparatedTokenizer()
    else:
        tokenizer = SpaceSeparatedTokenizer()
    if lowercase:
        tokenizer = tokenizer | LowercaseFilter()
    return tokenizer
KeywordAnalyzer.__inittypes__ = dict(lowercase=bool, commas=bool)


def RegexAnalyzer(expression=r"\w+(\.?\w+)*", gaps=False):
    """Deprecated, just use a RegexTokenizer directly.
    """
    
    return RegexTokenizer(expression=expression, gaps=gaps)
RegexAnalyzer.__inittypes__ = dict(expression=unicode, gaps=bool)


def SimpleAnalyzer(expression=r"\w+(\.?\w+)*", gaps=False):
    """Composes a RegexTokenizer with a LowercaseFilter.
    
    >>> ana = SimpleAnalyzer()
    >>> [token.text for token in ana(u"Hello there, this is a TEST")]
    [u"hello", u"there", u"this", u"is", u"a", u"test"]
    
    :param expression: The regular expression pattern to use to extract tokens.
    :param gaps: If True, the tokenizer *splits* on the expression, rather
        than matching on the expression.
    """
    
    return RegexTokenizer(expression=expression, gaps=gaps) | LowercaseFilter()
SimpleAnalyzer.__inittypes__ = dict(expression=unicode, gaps=bool)

def StandardAnalyzer(expression=r"\w+(\.?\w+)*", stoplist = STOP_WORDS, minsize = 2, gaps=False):
    """Composes a RegexTokenizer with a LowercaseFilter and optional StopFilter.
    
    >>> ana = StandardAnalyzer()
    >>> [token.text for token in ana(u"Testing is testing and testing")]
    [u"testing", u"testing", u"testing"]
    
    :param expression: The regular expression pattern to use to extract tokens.
    :param stoplist: A list of stop words. Set this to None to disable
        the stop word filter.
    :param minsize: Words smaller than this are removed from the stream.
    :param gaps: If True, the tokenizer *splits* on the expression, rather
        than matching on the expression.
    """
    
    chain = RegexTokenizer(expression=expression, gaps=gaps) | LowercaseFilter()
    if stoplist is not None:
        chain = chain | StopFilter(stoplist = stoplist, minsize = minsize)
    return chain
StandardAnalyzer.__inittypes__ = dict(expression=unicode, gaps=bool, stoplist=list, minsize=int)


def StemmingAnalyzer(expression=r"\w+(\.?\w+)*", stoplist=STOP_WORDS, minsize=2, gaps=False):
    """Composes a RegexTokenizer with a lower case filter, an optional stop filter,
    and a stemming filter.
    
    >>> ana = StemmingAnalyzer()
    >>> [token.text for token in ana(u"Testing is testing and testing")]
    [u"test", u"test", u"test"]
    
    :param expression: The regular expression pattern to use to extract tokens.
    :param stoplist: A list of stop words. Set this to None to disable
        the stop word filter.
    :param minsize: Words smaller than this are removed from the stream.
    :param gaps: If True, the tokenizer *splits* on the expression, rather
        than matching on the expression.
    """
    
    chain = RegexTokenizer(expression=expression, gaps=gaps) | LowercaseFilter()
    if stoplist is not None:
        chain = chain | StopFilter(stoplist = stoplist, minsize = minsize)
    return chain | StemFilter()
StemmingAnalyzer.__inittypes__ = dict(expression=unicode, gaps=bool, stoplist=list, minsize=int)


def FancyAnalyzer(expression=r"\s+", stoplist = STOP_WORDS, minsize = 2, gaps=True,
                  splitwords=True, splitnums=True,
                  mergewords=False, mergenums=False):
    """Composes a RegexTokenizer with a CamelFilter, UnderscoreFilter, LowercaseFilter,
    and StopFilter.
    
    >>> ana = FancyAnalyzer()
    >>> [token.text for token in ana(u"Should I call getInt or get_real?")]
    [u"should", u"call", u"getInt", u"get", u"int", u"get_real", u"get", u"real"]
    
    :param expression: The regular expression pattern to use to extract tokens.
    :param stoplist: A list of stop words. Set this to None to disable
        the stop word filter.
    :param minsize: Words smaller than this are removed from the stream.
    :param gaps: If True, the tokenizer *splits* on the expression, rather
        than matching on the expression.
    """
    
    return RegexTokenizer(expression=expression, gaps=gaps) | \
           IntraWordFilter(splitwords=splitwords, splitnums=splitnums,
                           mergewords=mergewords, mergenums=mergenums) | \
           LowercaseFilter() | \
           StopFilter(stoplist = stoplist, minsize = minsize)
FancyAnalyzer.__inittypes__ = dict(expression=unicode, gaps=bool, stoplist=list, minsize=int)


def NgramAnalyzer(minsize, maxsize = None):
    """Composes an NgramTokenizer and a LowercaseFilter.
    
    >>> ana = NgramAnalyzer(4)
    >>> [token.text for token in ana(u"hi there")]
    [u"hi t", u"i th", u" the", u"ther", u"here"]
    """
    
    return NgramTokenizer(minsize, maxsize = maxsize) | LowercaseFilter()
NgramAnalyzer.__inittypes__ = dict(minsize=int, maxsize=int)


    






#===============================================================================
# Copyright 2009 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""
This module contains classes for writing and reading postings.

The PostIterator interface is the base interface for the two "cursor" interfaces
(PostingReader and QueryScorer). It defines the basic methods for moving through
the posting list (e.g. ``reset()``, ``next()``, ``skip_to()``).

The PostingReader interface allows reading raw posting information.
Individual backends must provide a PostingReader implementation that will be returned
by the backend's :meth:`whoosh.reading.IndexReader.postings` method.
PostingReader subclasses in this module provide synthetic readers or readers that wrap
other readers and modify their behavior.

The QueryScorer interface allows retrieving and scoring search results. QueryScorer
objects will be returned by the :meth:`~whoosh.query.Query.scorer` method on
:class:`whoosh.query.Query` objects.
QueryScorer subclasses in this module provide synthetic scorers or scorers that wrap
other scorers and modify their behavior.
"""


from heapq import heapify, heappop, heapreplace


# Exceptions

class ReadTooFar(Exception):
    """Raised if a user calls next() or skip_to() on a reader that
    has reached the end of its items.
    """
    pass


# Base classes

class PostingWriter(object):
    def write(self, id, value):
        """Write the given id and value to the posting store.
        
        :param id: The identifier for this posting.
        :param value: The encoded value string for this posting.
        """
        raise NotImplementedError
    
    def finish(self):
        "Called when the current set of postings is finished."
        pass
    
    def close(self):
        raise NotImplementedError


class PostIterator(object):
    """Base class for PostingReader and QueryScorer. This interface
    provides methods for moving the "cursor" along posting lists.
    """
    
    def __cmp__(self, other):
        return cmp(self.id, other.id)
    
    def __repr__(self):
        return "<%s : %s>" % (self.__class__.__name__, self.id)
    
    def reset(self):
        "Resets the reader to the beginning of the postings"
        raise NotImplementedError(self)
    
    def next(self):
        "Moves to the next posting."
        raise NotImplementedError(self)
    
    def skip_to(self, id):
        """Skips ahead to the given id. The default implementation
        simply calls next() repeatedly until it gets to the id, but
        subclasses will often be more clever.
        """
        
        if id <= self.id:
            return
        if self.id is None:
            raise ReadTooFar
        
        next = self.next
        while self.id < id:
            next()
    
    # Iterator convenience functions
    
    def all_ids(self):
        """Yields all posting IDs.
        This may or may not change the cursor position, depending on the subclass
        and backend implementations.
        """
        self.reset()
        return self.ids()
    
    def ids(self):
        """Yields the remaining IDs in the reader.
        This may or may not change the cursor position, depending on the subclass
        and backend implementations.
        """
        
        next = self.next
        while self.id is not None:
            yield self.id
            next()


class PostingReader(PostIterator):
    """Base class for posting readers.
    
    "Postings" are used for two purposes in Whoosh.
    
    For each term in the index, the postings are the list of
    documents the term appears in and any associated value for each
    document. For example, if the field format is Frequency, the
    postings for the field might look like::
      
        [(0, 1), (10, 3), (12, 5)]
        
    ...where 0, 10, and 12 are document numbers, and 1, 3, and
    5 are the frequencies of the term in those documents.
      
    To get a PostingReader object for a term, use the
    :meth:`~whoosh.reading.IndexReader.postings` method on an IndexReader
    or Searcher object.
    
    >>> # Get a PostingReader for the term "render" in the "content" field.
    >>> r = myindex.reader()
    >>> preader = r.postings("content", u"render")
      
    For fields with term vectors, the vector postings are the list of
    terms that appear in the field and any associated value for each
    term. For example, if the term vector format is Frequency, the
    postings for the term vector might look like::
    
        [(u"apple", 1), (u"bear", 5), (u"cab", 2)]
        
    ...where "apple", "bear", and "cab" are the terms in the document
    field, and 1, 5, 2 are the frequencies of those terms in the document
    field.
    
    To get a PostingReader object for a vector, use the
    :meth:`~whoosh.reading.IndexReader.vector` method on an IndexReader
    or Searcher object.
    
    >>> # Get a PostingReader for the vector of the "content" field
    >>> # of document 100 
    >>> r = myindex.reader()
    >>> vreader = r.vector(100, "content")
    
    PostingReader defines a fairly simple interface.
    
    * The current posting ID is in the reader.id attribute.
    * Reader.value() to get the posting payload.
    * Reader.value_as(astype) to get the interpreted posting payload.
    * Reader.next() to move the reader to the next posting.
    * Reader.skip_to(id) to move the reader to that id in the list.
    * Reader.reset() to reset the reader to the beginning.
    
    In addition, PostingReader supports a few convenience methods:
    
    * ids() returns an iterator of the remaining IDs.
    * items() returns an iterator of the remaining (id, encoded_value) pairs.
    * items_as(astype) returns an interator of the remaining (id, decoded_value) pairs.
    
    all_ids(), all_items(), and all_as() are similar, but return iterators of
    *all* IDs/items in the reader, regardless of the current position of the reader.
      
    Different implementations may leave the reader in different positions during and
    after use of the iteration methods; that is, the effect of the iterators on the
    reader's position is undefined and may be different in different PostingReader
    subclasses and different backend implementations.
    """
    
    def value(self):
        "Returns the encoded value string for the current id."
        raise NotImplementedError

    def value_as(self, astype):
        """Returns the value for the current id as the given type.
        
        :param astype: a string, such as "weight" or "positions". The
            Format object associated with this reader must have a
            corresponding "as_*" method, e.g. as_weight(), for decoding
            the value.
        """
        return self.format.decoder(astype)(self.value())
    
    # Iterator convenience functions
    
    def all_items(self):
        """Yields all (id, encoded_value) pairs in the reader.
        Use all_as() to get decoded values.
        This may or may not change the cursor position, depending on the subclass
        and backend implementations.
        """
        self.reset()
        return self.items()
    
    def all_as(self, astype):
        """Yield a series of (id, decoded_value) pairs for each posting.
        This may or may not change the cursor position, depending on the subclass
        and backend implementations.
        """
        self.reset()
        return self.items_as(astype)
    
    def items(self):
        """Yields the remaining (id, encoded_value) pairs in the reader.
        Use items_as() to get decoded values.
        This may or may not change the cursor position, depending on the subclass
        and backend implementations.
        """
        
        next = self.next
        while self.id is not None:
            yield self.id, self.value()
            next()
            
    def items_as(self, astype):
        """Yields the remaining (id, decoded_value) pairs in the reader.
        This may or may not change the cursor position, depending on the subclass
        and backend implementations.
        """
        decoder = self.format.decoder(astype)
        for id, valuestring in self.items():
            yield (id, decoder(valuestring))


class QueryScorer(PostIterator):
    """QueryScorer extends the PostIterator interface with two methods:
    
    * score() return the score for the current item.
    * __iter__() returns an iterator of (id, score) pairs.
    """
    
    def __iter__(self):
        next = self.next
        while self.id is not None:
            yield self.id, self.score()
            next()
    
    def score(self):
        """Returns the score for the current document.
        """
        raise NotImplementedError


class FakeIterator(object):
    """A mix-in that provides methods for a fake PostingReader or
    QueryScorer.
    """
    
    def __init__(self, *ids):
        self.ids = ids
        self.reset()
        
    def reset(self):
        self.i = 0
        if self.ids:
            self.id = self.ids[0]
        else:
            self.id = None
    
    def next(self):
        if self.id is None:
            raise ReadTooFar
        
        if self.i == len(self.ids) - 1:
            self.id = None
        else:
            self.i += 1
            self.id = self.ids[self.i]
    
    def skip_to(self, target):
        if target <= self.id:
            return
        if self.id is None:
            raise ReadTooFar
        
        i, ids = self.i, self.ids
        
        while ids[i] < target:
            i += 1
            if i == len(ids):
                self.id = None
                return
        
        self.i = i
        self.id = ids[i]


# Posting readers

class MultiPostingReader(PostingReader):
    """This posting reader concatenates the results from serial sub-readers.
    This is useful for backends that use a segmented index.
    """
    
    def __init__(self, format, readers, idoffsets):
        """
        :param format: the :class:`~whoosh.formats.Format` object for the field
            being read.
        :param readers: a list of :class:`~whoosh.postings.PostingReader` objects.
        :param idoffsets: a list of integers, where each item in the list represents
            the ID offset of the corresponding reader in the 'readers' list.
        """
        
        self.format = format
        self.readers = readers
        self.offsets = idoffsets
        self.current = 0
        self._prep()
    
    def _prep(self):
        readers = self.readers
        current = self.current
        if not readers:
            self.id = None
            
        while readers[current].id is None:
            current += 1
            if current >= len(readers):
                self.id = None
                return
        
        self.current = current
        self.id = readers[current].id + self.offsets[current]

    def reset(self):
        if not self.readers:
            return
        
        for r in self.readers:
            r.reset()
        self.current = 0
        self._prep()

    def all_items(self):
        offsets = self.offsets
        for i, r in enumerate(self.readers):
            for id, valuestring in r.all_items():
                yield id + offsets[i], valuestring

    def all_ids(self):
        offsets = self.offsets
        for i, r in enumerate(self.readers):
            for id in r.all_ids():
                yield id + offsets[i]

    def next(self):
        if self.id is None:
            raise ReadTooFar
        
        readers = self.readers
        current = self.current
        readers[current].next()
        while current < len(readers) - 1 and self.readers[current].id is None:
            current += 1
            
        if self.readers[current].id is None:
            self.id = None
        else:
            self.id = readers[current].id + self.offsets[current]
            self.current = current

    def skip_to(self, target):
        if target <= self.id:
            return
        if self.id is None:
            raise ReadTooFar
        
        current = self.current
        readers = self.readers
        offsets = self.offsets
        
        while current < len(readers):
            r = readers[current]
            if r.id is None:
                current += 1
                continue
            
            if target < r.id:
                self.current = current
                self.id = r.id + offsets[current]
                return
            
            r.skip_to(target - offsets[current])
            if r.id is not None:
                self.current = current
                self.id = r.id + offsets[current]
                return
            
            current += 1
            
        self.id = None

    def value(self):
        return self.readers[self.current].value()


class Exclude(PostingReader):
    """PostingReader that removes certain IDs from a sub-reader.
    """
    
    def __init__(self, postreader, excludes):
        """
        :param postreader: the PostingReader object to read from.
        :param excludes: a collection of ids to exclude (may be any object,
            such as a BitVector or set, that implements __contains__).
        """
        
        self.postreader = postreader
        if hasattr(postreader, "format"):
            self.format = postreader.format
        
        self.excludes = excludes
        self._find_nonexcluded()
        self.value = postreader.value
    
    def reset(self):
        self.postreader.reset()
        self._find_nonexcluded()
    
    def _find_nonexcluded(self):
        pr, excl = self.postreader, self.excludes
        next = pr.next
        while pr.id is not None and pr.id in excl:
            next()
        self.id = pr.id
    
    def next(self):
        self.postreader.next()
        self._find_nonexcluded()
    
    def skip_to(self, target):
        if target <= self.id:
            return
        self.postreader.skip_to(target)
        self._find_nonexcluded()


class CachedPostingReader(PostingReader):
    """Reads postings from a list in memory instead of from storage.
    
    >>> preader = ixreader.postings("content", "render")
    >>> creader = CachedPostingReader(preader.all_items())
    """
    
    def __init__(self, items):
        """
        :param items: a sequence of (id, encodedvalue) pairs. If this is
            not a list or tuple, it is converted using tuple().
        """
        
        if not isinstance(items, (list, tuple)):
            items = tuple(items)
        
        self._items = items
        self.reset()
    
    def reset(self):
        self.p = 0
        self.id = self._items[0][0]
    
    def all_ids(self):
        return (item[0] for item in self._items)
    
    def all_items(self):
        return iter(self._items)
    
    def next(self):
        if self.id is None:
            raise ReadTooFar
        
        self.p += 1
        if self.p >= len(self._items):
            self.id = None
        else:
            self.id = self._items[self.p][0]
            
    def skip_to(self, target):
        if self.id is None:
            raise ReadTooFar
        if target < self.id:
            return
        
        items = self._items
        p = self.p + 1
        while p < len(items):
            id = items[p][0]
            if id >= target:
                self.p = p
                self.id = id
                return
        
        self.id = None
        
    def value(self):
        return self._items[self.p][1]


class FakeReader(FakeIterator, PostingReader):
    """This is a fake posting reader for testing purposes. You create the
    object with the posting IDs as arguments, and then returns them as you
    call next() or skip_to().
    
    >>> fpr = FakeReader(1, 5, 10, 80)
    >>> fpr.id
    1
    >>> fpr.next()
    >>> fpr.id
    5
    """
    
    _value = 100
    
    def value(self):
        return self._value
    

# QueryScorers

class FakeScorer(FakeIterator, QueryScorer):
    """This is a fake query scorer for testing purposes. You create the
    object with the posting IDs as arguments, and then returns them as you
    call next() or skip_to().
    
    >>> fpr = FakeScorer(1, 5, 10, 80)
    >>> fpr.id
    1
    >>> fpr.next()
    >>> fpr.id
    5
    """
    
    _score = 10
    
    def score(self):
        return self._score


class EmptyScorer(QueryScorer):
    """A QueryScorer representing a query that doesn't match any documents.
    """
    
    def __init__(self):
        self.id = None
    def reset(self):
        pass
    def next(self):
        pass
    def skip_to(self, id):
        pass
    def ids(self):
        return []
    def items(self):
        return []
    def items_as(self, astype):
        return []
    def score(self):
        return 0


class IntersectionScorer(QueryScorer):
    """Acts like the intersection of items in a set of QueryScorers
    """
    
    def __init__(self, scorers, boost=1.0):
        self.scorers = scorers
        self.state = list(scorers)
        self.boost = boost
        self.id = -1
        self._prep()

    def __repr__(self):
        return "<%s %r: %r>" % (self.__class__.__name__, self.scorers, self.id)

    def _prep(self):
        state = self.state
        state.sort()
        id = state[0].id
        if all(r.id == id for r in state[1:]):
            self.id = id
        else:
            self.next()

    def reset(self):
        for r in self.state:
            r.reset()
        self.id = -1
        self._prep()

    def skip_to(self, target):
        if self.id is None:
            raise ReadTooFar
        
        state = self.state
        for r in state:
            r.skip_to(target)
        self._prep()
    
    def next(self):
        if self.id is None:
            raise ReadTooFar
        
        id = self.id
        state = self.state
        for r in state:
            if r.id == id: r.next()
        state.sort()
        
        while True:
            oldstate = tuple(s.id for s in state)
            lowid = state[0].id
            if lowid is None:
                self.id = None
                return
            
            if all(r.id == lowid for r in state[1:]):
                self.id = lowid
                return
            else:
                highid = state[-1].id
                for r in state[:-1]:
                    r.skip_to(highid)
                if state[0].id is not None:
                    state.sort()
            if tuple(s.id for s in state) == oldstate:
                raise Exception("Infinite loop")
                    
    def score(self):
        if self.id is None:
            return 0
        return sum(r.score() for r in self.state) * self.boost
                

class UnionScorer(QueryScorer):
    """Acts like the union of a set of QueryScorers
    """
    
    def __init__(self, scorers, boost=1.0):
        self.scorers = scorers
        self.boost = boost
        self.state = [s for s in scorers if s.id is not None]
        
        if self.state:
            heapify(self.state)
            self.id = self.state[0].id
        else:
            self.id = None
    
    def reset(self):
        for s in self.scorers:
            s.reset()
        self.state = [s for s in self.scorers if s.id is not None]
        
        if self.state:
            heapify(self.state)
            self.id = self.state[0].id
        else:
            self.id = None
    
    def skip_to(self, target):
        if self.id is None:
            raise ReadTooFar
        
        state = self.state
        for r in state:
            r.skip_to(target)
        
        heapify(state)
        while state and state[0].id is None:
            heappop(state)
        
        if state:
            self.id = self.state[0].id
        else:
            self.id = None
        
    def next(self):
        if self.id is None:
            raise ReadTooFar
        
        state = self.state
        
        # Short circuit if there's only one reader
        if len(state) == 1:
            r = state[0]
            r.next()
            self.id = r.id
        else:
            lowid = state[0].id
            while state and state[0].id == lowid:
                r = state[0]
                r.next()
                if r.id is None:
                    heappop(state)
                else:
                    heapreplace(state, r)
        
            if state:
                self.id = state[0].id
            else:
                self.id = None
            
    def score(self):
        id = self.id
        if id is None:
            return 0
        score = sum(r.score() for r in self.state if r.id == id)
        return score * self.boost


class AndNotScorer(QueryScorer):
    """Takes two QueryScorers and pulls items from the first,
    removing ids that also appear in the second.
    """
    
    def __init__(self, positive, negative):
        """
        :param positive: a QueryScorer from which to take items.
        :param negative: a QueryScorer, the IDs of which will be
            removed from the 'positive' scorer.
        """
        
        self.positive = positive
        self.negative = negative
        self._find_next()
    
    def reset(self):
        self.positive.reset()
        self.negative.reset()
        self._find_next()
    
    def _find_next(self):
        inp, outp = self.positive, self.negative
        while inp.id is not None and inp.id == outp.id:
            inp.next()
            if outp.id is not None:
                outp.skip_to(inp.id)
        self.id = inp.id
    
    def next(self):
        if self.id is None:
            raise ReadTooFar
        
        self.positive.next()
        if self.negative.id:
            self._find_next()
        else:
            self.id = self.positive.id
        
    def skip_to(self, target):
        if self.id is None:
            raise ReadTooFar
        if target <= self.id:
            return
        
        self.positive.skip_to(target)
        if self.negative.id is not None:
            self.negative.skip_to(target)
            self._find_next()
        else:
            self.id = self.positive.id
        
    def score(self):
        if self.id is None:
            return 0
        return self.positive.score()


class RequireScorer(QueryScorer):
    """Takes the intersection of two sub-scorers, but only
    takes scores from the first.
    """
    
    def __init__(self, scorer, required):
        self.scorer = scorer
        self.intersection = IntersectionScorer([scorer, required])
        
        self.reset = self.intersection.reset
        self.next = self.intersection.next
        self.skip_to = self.intersection.skip_to
    
    @property
    def id(self):
        return self.intersection.id
    
    def score(self):
        if self.id is None:
            return 0
        return self.scorer.score()


class AndMaybeScorer(QueryScorer):
    """Takes two sub-scorers, and returns documents that appear in the
    first, but if the document also appears in the second, adds their
    scores together.
    """
    
    def __init__(self, required, optional):
        self.required = required
        self.optional = optional
        
    def reset(self):
        self.required.reset()
        self.optional.reset()
        
    def next(self):
        self.required.next()
        self.optional.skip_to(self.required.id)

    def skip_to(self, target):
        self.required.skip_to(target)
        self.optional.skip_to(target)

    def score(self):
        if self.required.id == self.optional.id:
            return self.required.score() + self.optional.score()
        else:
            return self.required.score()














#===============================================================================
# Copyright 2007 Matt Chaput
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#    http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#===============================================================================

"""
Miscellaneous utility functions and classes.
"""

import re

from collections import deque, defaultdict
from functools import wraps
from struct import pack, unpack
from time import time, clock

# Functions

# Varint cache

# Build a cache of the varint byte sequences for the first
# N integers, so we don't have to constantly recalculate them
# on the fly. This makes a small but noticeable difference.

def _varint(i):
    s = ""
    while (i & ~0x7F) != 0:
        s += chr((i & 0x7F) | 0x80)
        i = i >> 7
    s += chr(i)
    return s

_varint_cache_size = 512
_varint_cache = []
for i in xrange(0, _varint_cache_size):
    _varint_cache.append(_varint(i))
_varint_cache = tuple(_varint_cache)

def varint(i):
    """Encodes the given integer into a string of the minimum number
    of bytes.
    """
    if i < len(_varint_cache):
        return _varint_cache[i]
    return _varint(i)

def read_varint(readfn):
    """
    Reads a variable-length encoded integer.
    
    :param readfn: a callable that reads a given number of bytes,
        like file.read().
    """
    
    b = ord(readfn(1))
    i = b & 0x7F

    shift = 7
    while b & 0x80 != 0:
        b = ord(readfn(1))
        i |= (b & 0x7F) << shift
        shift += 7
    return i


_fib_cache = {}
def fib(n):
    """Returns the nth value in the Fibonacci sequence."""
    
    if n <= 2: return n
    if n in _fib_cache: return _fib_cache[n]
    result = fib(n - 1) + fib(n - 2)
    _fib_cache[n] = result
    return result


def float_to_byte(value, mantissabits = 5, zeroexp = 2):
    """Encodes a floating point number in a single byte.
    """
    
    # Assume int size == float size
    
    fzero = (63 - zeroexp) << mantissabits
    bits = unpack("i", pack("f", value))[0]
    smallfloat = bits >> (24 - mantissabits)
    if smallfloat < fzero:
        # Map negative numbers and 0 to 0
        # Map underflow to next smallest non-zero number
        if bits <= 0:
            return chr(0)
        else:
            return chr(1)
    elif smallfloat >= fzero + 0x100:
        # Map overflow to largest number
        return chr(255)
    else:
        return chr(smallfloat - fzero)
    
def byte_to_float(b, mantissabits = 5, zeroexp = 2):
    """Decodes a floating point number stored in a single
    byte.
    """
    b = ord(b)
    if b == 0:
        return 0.0
    
    bits = (b & 0xff) << (24 - mantissabits)
    bits += (63 - zeroexp) << 24
    return unpack("f", pack("i", bits))[0]


def first_diff(a, b):
    """Returns the position of the first differing character in the strings
    a and b. For example, first_diff('render', 'rending') == 4. This
    function limits the return value to 255 so the difference can be encoded
    in a single byte.
    """
    
    i = -1
    for i in xrange(0, len(a)):
        if a[i] != b[1]:
            return i
        if i == 255: return i


def prefix_encode(a, b):
    """Compresses string b as an integer (encoded in a byte) representing
    the prefix it shares with a, followed by the suffix encoded as UTF-8.
    """
    i = first_diff(a, b)
    return chr(i) + b[i:].encode("utf8")


def prefix_encode_all(ls):
    """Compresses the given list of (unicode) strings by storing each string
    (except the first one) as an integer (encoded in a byte) representing
    the prefix it shares with its predecessor, followed by the suffix encoded
    as UTF-8.
    """
    
    last = u''
    for w in ls:
        i = first_diff(last, w)
        yield chr(i) + w[i:].encode("utf8")
        last = w
        

def prefix_decode_all(ls):
    """Decompresses a list of strings compressed by prefix_encode().
    """
    
    last = u''
    for w in ls:
        i = ord(w[0])
        decoded = last[:i] + w[1:].decode("utf8")
        yield decoded
        last = decoded


_nkre = re.compile(r"\D+|\d+", re.UNICODE)
def _nkconv(i):
    try:
        return int(i)
    except ValueError:
        return i.lower()
def natural_key(s):
    """Converts string ``s`` into a tuple that will sort "naturally"
    (i.e., ``name5`` will come before ``name10`` and ``1`` will come
    before ``A``). This function is designed to be used as the ``key``
    argument to sorting functions.
    
    :param s: the str/unicode string to convert.
    :rtype: tuple
    """
    
    # Use _nkre to split the input string into a sequence of
    # digit runs and non-digit runs. Then use _nkconv() to convert
    # the digit runs into ints and the non-digit runs to lowercase.
    return tuple(_nkconv(m) for m in _nkre.findall(s))


class ClosableMixin(object):
    """Mix-in for classes with a close() method to allow them to be
    used as a context manager.
    """
    
    def __enter__(self):
        return self
    
    def __exit__(self, *exc_info):
        self.close()


def protected(func):
    """Decorator for storage-access methods. This decorator
    (a) checks if the object has already been closed, and
    (b) synchronizes on a threading lock. The parent object must
    have 'is_closed' and '_sync_lock' attributes.
    """
    
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.is_closed:
            raise Exception("%r has been closed" % self)
        if self._sync_lock.acquire(False):
            try:
                return func(self, *args, **kwargs)
            finally:
                self._sync_lock.release()
        else:
            raise Exception("Could not acquire sync lock")
    
    return wrapper


def lru_cache(size):
    """Decorator that adds a least-recently-accessed cache to a method.
    
    :param size: the maximum number of items to keep in the cache.
    """
    
    def decorate_function(func):
        prefix = "_%s_" % func.__name__
    
        @wraps(func)
        def wrapper(self, *args):
            if not hasattr(self, prefix + "cache"):
                cache = {}
                queue = deque()
                refcount = defaultdict(int)
                setattr(self, prefix + "cache", cache)
                setattr(self, prefix + "queue", queue)
                setattr(self, prefix + "refcount", refcount)
            else:
                cache = getattr(self, prefix + "cache")
                queue = getattr(self, prefix + "queue")
                refcount = getattr(self, prefix + "refcount")
            qpend = queue.append
            qpop = queue.popleft
            
            # Get cache entry or compute if not found
            try:
                result = cache[args]
            except KeyError:
                result = cache[args] = func(self, *args)
    
            # Record that this key was recently accessed
            qpend(args)
            refcount[args] += 1
    
            # Purge least recently accessed cache contents
            while len(cache) > size:
                k = qpop()
                refcount[k] -= 1
                if not refcount[k]:
                    del cache[k]
                    del refcount[k]
    
            # Periodically compact the queue by removing duplicate keys
            if len(queue) > size * 4:
                for _ in xrange(len(queue)):
                    k = qpop()
                    if refcount[k] == 1:
                        qpend(k)
                    else:
                        refcount[k] -= 1
                #assert len(queue) == len(cache) == len(refcount) == sum(refcount.itervalues())
            
            return result
        return wrapper
    return decorate_function




















